import requests
from typing import Type, Union

from peewee import Model as PeeweeModel, DoesNotExist, fn
from loguru import logger

ID = Union[str, int]


class Client:
    """Client to communicate with the API"""

    def __init__(self, endpoint: str) -> None:
        self.endpoint = endpoint

    def create(self, record: dict):
        record = self.filter_none(record)
        assert "id" not in record
        resp = requests.post(f"{self.endpoint}", json=record)
        self.assert_resp(resp)
        return resp.json()

    def upsert(self, record: dict):
        record = self.filter_none(record)

        if "id" in record:
            self.update(record)
            return record

        resp = requests.get(self.endpoint, params=record)
        self.assert_resp(resp)
        items = resp.json()["items"]
        if len(items) == 0:
            # not found, we create it
            resp = requests.post(self.endpoint, json=record)
            self.assert_resp(resp)
            record = resp.json()
        else:
            assert len(items) == 1, "Upsert should only affect one record"
            record = items[0]

        return record

    def update(self, record: dict):
        resp = requests.put(
            f"{self.endpoint}/{record['id']}", json=self.filter_none(record)
        )
        self.assert_resp(resp)
        return resp.json()

    def has(self, record_or_id: Union[dict, ID]):
        if (
            not isinstance(record_or_id, dict)
            or record_or_id.get("id", None) is not None
        ):
            id = record_or_id["id"] if isinstance(record_or_id, dict) else record_or_id

            resp = requests.head(f"{self.endpoint}/{id}")
            if resp.status_code == 404:
                return False
            self.assert_resp(resp)
            return True

        record = self.filter_none(record_or_id)
        resp = requests.get(f"{self.endpoint}", params=record)
        self.assert_resp(resp)
        return len(resp.json()["items"]) > 0

    def get(self, queries: dict):
        resp = requests.get(f"{self.endpoint}", params=queries)
        self.assert_resp(resp)
        return resp.json()["items"]

    def get_one(self, queries: dict):
        items = self.get(queries)
        if len(items) == 0:
            raise Exception("No item found")
        assert len(items) == 1, items
        return items[0]

    def get_by_id(self, id: str):
        resp = requests.get(f"{self.endpoint}/{id}")
        self.assert_resp(resp)
        return resp.json()

    def delete(self, record_or_id: Union[dict, ID]):
        if (
            not isinstance(record_or_id, dict)
            or record_or_id.get("id", None) is not None
        ):
            id = record_or_id["id"] if isinstance(record_or_id, dict) else record_or_id
            resp = requests.delete(f"{self.endpoint}/{id}")
            self.assert_resp(resp)
        else:
            raise NotImplementedError("The function does not support on server as well")

    def filter_none(self, record: dict):
        return {k: v for k, v in record.items() if v is not None}

    def assert_resp(self, resp, status_code: int = 200):
        if resp.status_code != status_code:
            logger.error(f"Expect status code {status_code} but get {resp.status_code}")
            logger.error(resp.text)
            raise Exception(resp.status_code)



from __future__ import annotations
import orjson
from typing import Type
from peewee import Field
from dataclasses import astuple


# db_null = b"null"
db_null = None


class DataClassField(Field):
    field_type = "BLOB"

    def __init__(self, CLS: Type[object], **kwargs):
        super().__init__(**kwargs)
        self.CLS = CLS
        if hasattr(CLS, "from_tuple"):
            self.from_tuple = getattr(CLS, "from_tuple")
        else:
            self.from_tuple = lambda x: CLS(*x)  # type: ignore

    def db_value(self, value):
        if value is None:
            return value
        return orjson.dumps(astuple(value))

    def python_value(self, value):
        if value == db_null:
            return None
        value = orjson.loads(value)
        return self.from_tuple(value)


class ListDataClassField(DataClassField):
    def db_value(self, value):
        return orjson.dumps([astuple(item) for item in value])

    def python_value(self, value):
        if value == db_null:
            return None

        value = orjson.loads(value)
        return [self.from_tuple(item) for item in value]


class DictDataClassField(DataClassField):
    def db_value(self, value):
        return orjson.dumps({k: astuple(item) for k, item in value.items()})

    def python_value(self, value):
        if value == db_null:
            return None

        value = orjson.loads(value)
        return {k: self.from_tuple(item) for k, item in value.items()}


class Dict2ListDataClassField(DataClassField):
    def db_value(self, value):
        try:
            return orjson.dumps(
                {k: [astuple(item) for item in lst] for k, lst in value.items()}
            )
        except:
            print(value)
            raise

    def python_value(self, value):
        if value == db_null:
            return None

        value = orjson.loads(value)
        return {k: [self.from_tuple(item) for item in lst] for k, lst in value.items()}



from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    _TypedDictMeta,  # type: ignore
)
from gena.custom_fields import (
    DataClassField,
    Dict2ListDataClassField,
    DictDataClassField,
    ListDataClassField,
)
from peewee import (
    CharField,
    Model,
    Field,
    FloatField,
    ForeignKeyField,
    IntegerField,
    BooleanField,
    _StringField,
    TextField,
    Value,
)
from dataclasses import fields, is_dataclass


class NoDerivedDeserializer(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error_trace = []

    def get_root_field(self):
        return self.error_trace[-1]

    def add_trace(self, *parents: str):
        self.error_trace.extend(reversed(parents))
        return self

    def __str__(self) -> str:
        return f"cannot derive deserializer for: {list(reversed(self.error_trace))}"


Deserializer = Callable[[Any], Any]


def generate_deserializer(
    Model: Type[Model],
    known_type_deserializers: Optional[Dict[Any, Deserializer]] = None,
    known_field_deserializers: Optional[Set[str]] = None,
) -> Dict[str, Deserializer]:
    known_type_deserializers = known_type_deserializers or {}
    known_field_deserializers = known_field_deserializers or set()

    fields = Model._meta.fields
    field_type_hints = get_type_hints(Model)

    output = {}
    for name, field in fields.items():
        if name in known_field_deserializers:
            continue

        if isinstance(field, IntegerField):
            func = deserialize_int
        elif isinstance(field, FloatField):
            func = deserialize_float
        elif isinstance(field, BooleanField):
            func = deserialize_bool
        elif isinstance(field, _StringField):
            if isinstance(field, CharField) or type(field) is TextField:
                if (
                    name in field_type_hints
                    and get_origin(field_type_hints[name]) is Literal
                ):
                    func = get_deserializer_from_type(
                        field_type_hints[name], known_type_deserializers
                    )
                else:
                    func = deserialize_str
            else:
                # a catch here, JSONField is a subclass of _StringField & TextField
                # so we check if its type hint
                if name not in field_type_hints:
                    raise Exception(
                        f"Column {name} is not annotated or not a class variable. But we need its annotation because the field does not seems to be a string field"
                    )
                elif field_type_hints[name] is not str:
                    func = get_deserializer_from_type(
                        field_type_hints[name], known_type_deserializers
                    )
                else:
                    func = deserialize_str
        elif isinstance(field, ForeignKeyField):
            if field.field_type.upper() == "INT":
                func = deserialize_int
            else:
                raise NoDerivedDeserializer().add_trace(Model.__qualname__, name)
        elif isinstance(field, DataClassField):
            func = get_dataclass_deserializer(field.CLS, known_type_deserializers)
            if type(field) is ListDataClassField:
                func = get_deserialize_list(func)
            elif type(field) is DictDataClassField:
                func = get_deserialize_dict(func)
            elif type(field) is Dict2ListDataClassField:
                func = get_deserialize_dict(get_deserialize_list(func))
        else:
            raise NoDerivedDeserializer().add_trace(Model.__qualname__, name)

        if field.null:
            func = get_deserialize_nullable(func)
        output[name] = func
    return output


def deserialize_int(value):
    if isinstance(value, int):
        return value

    if isinstance(value, str):
        return int(value)

    if isinstance(value, float):
        if value == int(value):
            return int(value)

    raise ValueError(f"expect integer but get: {type(value)}")


def deserialize_bool(value):
    if isinstance(value, bool):
        return value

    if isinstance(value, str):
        if value != "true" and value != "false":
            raise ValueError(f"expect bool string but get: {value}")
        return value == "true"

    raise ValueError(f"expect bool value but get: {type(value)}")


def deserialize_str(value):
    if isinstance(value, str):
        return value
    raise ValueError(f"expect string but get: {type(value)}")


def deserialize_float(value):
    if isinstance(value, (int, float)):
        return value

    if isinstance(value, str):
        return float(value)

    raise ValueError(f"expect float but get: {type(value)}")


def deserialize_number_or_string(value):
    if not isinstance(value, (int, str, float)):
        raise ValueError(f"expect either string or number but get {type(value)}")
    return value


def deserialize_none(value):
    if value is not None:
        raise ValueError(f"expect none but get {type(value)}")
    return value


def get_deserialize_nullable(deserialize_item: Deserializer):
    def deserialize_nullable(value):
        if value is None:
            return None
        return deserialize_item(value)

    return deserialize_nullable


def get_deserialize_list(deserialize_item: Deserializer):
    def deserialize_list(value):
        if not isinstance(value, list):
            raise ValueError(f"expect list but get {type(value)}")
        return [deserialize_item(item) for item in value]

    return deserialize_list


def get_deserialize_homogeneous_tuple(deserialize_item: Deserializer):
    def deserialize_tuple(value):
        if not isinstance(value, (list, tuple)):
            raise ValueError(f"expect list/tuple but get {type(value)}")
        return tuple(deserialize_item(item) for item in value)

    return deserialize_tuple


def get_deserialize_set(deserialize_item: Deserializer):
    def deserialize_set(value):
        if not isinstance(value, set):
            raise ValueError(f"expect set but get {type(value)}")
        return {deserialize_item(item) for item in value}

    return deserialize_set


def get_deserialize_dict(deserialize_item: Deserializer):
    def deserialize_dict(value):
        if not isinstance(value, dict):
            raise ValueError(f"expect dict but get {type(value)}")
        return {k: deserialize_item(item) for k, item in value.items()}

    return deserialize_dict


def get_deserializer_from_type(
    annotated_type, known_type_deserializers: Dict[Any, Deserializer]
) -> Deserializer:
    if annotated_type in known_type_deserializers:
        return known_type_deserializers[annotated_type]
    if annotated_type is str:
        return deserialize_str
    if annotated_type is int:
        return deserialize_int
    if annotated_type is float:
        return deserialize_float
    if annotated_type is bool:
        return deserialize_bool
    if annotated_type is type(None):
        return deserialize_none
    if is_dataclass(annotated_type):
        return get_dataclass_deserializer(annotated_type, known_type_deserializers)
    if isinstance(annotated_type, _TypedDictMeta):
        # is_typeddict is not supported at python 3.8 yet
        return get_typeddict_deserializer(annotated_type, known_type_deserializers)
    try:
        if issubclass(annotated_type, Enum):
            # enum can be reconstructed using its constructor.
            return annotated_type
    except TypeError:
        pass

    args = get_args(annotated_type)
    origin = get_origin(annotated_type)

    if origin is None or len(args) == 0:
        # we can't handle this type, e.g., some class that are not dataclass, or simply just list or set (not enough information)
        raise NoDerivedDeserializer().add_trace(annotated_type)

    # handle literal first
    if origin is Literal:
        assert all(
            isinstance(arg, (str, int, float)) for arg in args
        ), f"Invalid literals: {args}"
        valid_values = set(args)

        def deserialize_literal(value):
            if value not in valid_values:
                raise Exception(f"expect one of {valid_values} but get {value}")
            return value

        return deserialize_literal

    # handle a special case of variable-length tuple of homogeneous type
    # https://docs.python.org/3/library/typing.html#typing.Tuple
    if origin is tuple and len(args) > 1 and args[-1] is Ellipsis:
        if len(args) != 2:
            raise Exception(
                "invalid annotation of variable-length tuple of homogeneous type. expect one type and ellipsis"
            )
        return get_deserialize_homogeneous_tuple(
            get_deserializer_from_type(args[0], known_type_deserializers)
        )

    arg_desers = [
        get_deserializer_from_type(arg, known_type_deserializers) for arg in args
    ]
    if any(fn is None for fn in arg_desers):
        raise NoDerivedDeserializer().add_trace(
            annotated_type,
            next((arg for fn, arg in zip(arg_desers, args) if fn is None)),
        )

    deserialize_args: Deserializer
    if len(arg_desers) == 1:
        deserialize_args = arg_desers[0]  # type: ignore
    elif len(arg_desers) == 2 and type(None) in args:
        # handle special case of none
        not_none_arg_deser = [
            arg_desers[i] for i, arg in enumerate(args) if arg is not type(None)
        ][0]

        def deserialize_optional_arg(value):
            if value is None:
                return value
            return not_none_arg_deser(value)  # type: ignore

        deserialize_args = deserialize_optional_arg
    else:
        # TODO: we can optimize this further
        def deserialize_n_args(value):
            for arg_deser in arg_desers:
                try:
                    return arg_deser(value)  # type: ignore
                except ValueError:
                    pass
            raise ValueError(
                f"Expect one of the type: {''.join(str(arg) for arg in args)} but get {value}"
            )

        deserialize_args = deserialize_n_args

    if origin is list:
        return get_deserialize_list(deserialize_args)

    if origin is set:
        return get_deserialize_set(deserialize_args)

    if origin is dict:
        return get_deserialize_dict(deserialize_args)

    if origin is Union:
        return deserialize_args

    # do we exhaust the list of built-in types?
    raise NoDerivedDeserializer().add_trace(annotated_type)


def get_typeddict_deserializer(
    typeddict: _TypedDictMeta, known_type_deserializers: Dict[str, Deserializer]
) -> Deserializer:
    total = typeddict.__total__
    if not total:
        # they can inject any key as the semantic of total
        raise NoDerivedDeserializer().add_trace(
            typeddict.__name__, "is not a total TypedDict"
        )

    field2deserializer = {}

    def deserialize_typed_dict(value):
        if not isinstance(value, dict):
            raise ValueError("expect dictionary but get {value}")
        output = {}
        for field, func in field2deserializer.items():
            if field not in value:
                raise ValueError(f"expect field {field} but it's missing")
            output[field] = func(value[field])
        return output

    # assign first to support recursive type in the field
    known_type_deserializers[typeddict] = deserialize_typed_dict

    for field, field_type in typeddict.__annotations__.items():
        try:
            func = get_deserializer_from_type(field_type, known_type_deserializers)
        except NoDerivedDeserializer as e:
            del known_type_deserializers[typeddict]
            raise e.add_trace(field)
        field2deserializer[field] = func

    return deserialize_typed_dict


def get_dataclass_deserializer(
    CLS, known_type_deserializers: Dict[Any, Deserializer]
) -> Deserializer:
    # extract deserialize for each field
    field2deserializer: Dict[str, Deserializer] = {}
    field2optional: Dict[str, bool] = {}
    field_types = get_type_hints(CLS)

    def deserialize_dataclass(value):
        if not isinstance(value, dict):
            raise ValueError(f"expect dictionary but get {value}")

        output = {}
        for field, deserialize in field2deserializer.items():
            if field in value:
                output[field] = deserialize(value[field])
            elif not field2optional[field]:
                # not optional field but missing
                raise ValueError(f"expect the field {field} but it's missing")
        return CLS(**output)

    # assign first to support recursive type in the field
    known_type_deserializers[CLS] = deserialize_dataclass

    for field in fields(CLS):
        field_type = field_types[field.name]
        try:
            func = get_deserializer_from_type(field_type, known_type_deserializers)
        except NoDerivedDeserializer as e:
            # can't automatically figure out its child deserializer
            del known_type_deserializers[CLS]
            raise e.add_trace(CLS.__qualname__, field.name)

        field2deserializer[field.name] = func
        field2optional[field.name] = get_origin(field_type) is Union and type(
            None
        ) in get_args(field_type)

    return deserialize_dataclass


def deserialize_dict(value):
    """Deserialize a dictionary. Avoid using it because it does not deep check as other functions"""
    if not isinstance(value, dict):
        raise ValueError(f"expect dictionary but get {type(value)}")
    return value



from dataclasses import fields, is_dataclass
from datetime import datetime
from enum import Enum
from operator import attrgetter
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    _TypedDictMeta,  # type: ignore
)
from loguru import logger
import orjson
from gena.custom_fields import (
    DataClassField,
    Dict2ListDataClassField,
    DictDataClassField,
    ListDataClassField,
)
from peewee import (
    CharField,
    Model,
    Field,
    FloatField,
    ForeignKeyField,
    IntegerField,
    BooleanField,
    _StringField,
    DateField,
    DateTimeField,
    TimestampField,
    TextField,
    Value,
)

from gena.deserializer import get_deserialize_dict


class NoDerivedSerializer(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error_trace = []

    def add_trace(self, *parents: str):
        self.error_trace.extend(reversed(parents))
        return self

    def __str__(self) -> str:
        return f"cannot derive serializer for: {list(reversed(self.error_trace))}"


def datetime_serializer(value: Optional[datetime]):
    # return the number of milliseconds since the UNIX epoch
    if value is None:
        return None
    return round(value.timestamp() * 1000)


Serializer = Callable[[Any], Any]
default_known_type_serializer = {
    datetime: datetime_serializer,
}
NoneType = type(None)


def get_peewee_serializer(
    Model: Type[Model],
    known_type_serializer: Optional[Dict[Any, Serializer]] = None,
    exclude_fields: Optional[Set[str]] = None,
) -> Callable[[Model], dict]:
    known_type_serializer = known_type_serializer or {}
    exclude_fields = exclude_fields or set()
    fields = Model._meta.fields

    for k, v in default_known_type_serializer.items():
        if k not in known_type_serializer:
            known_type_serializer[k] = v

    field2func = {}
    foreign_keys = []
    for name, field in fields.items():
        if name in exclude_fields:
            continue

        func = None
        if isinstance(
            field,
            (IntegerField, FloatField, BooleanField, _StringField),
        ):
            # event if it is JSON, we keep the raw value (which is dictionary -- python value)
            func = None
        elif isinstance(field, ForeignKeyField):
            # handle foreign key separately to avoid extra query whenever we use the
            # proxy object
            foreign_keys.append([name, f"{name}_id"])
            continue
        elif isinstance(field, DataClassField):
            try:
                func = get_dataclass_serializer(field.CLS, known_type_serializer)
            except NoDerivedSerializer as e:
                raise e.add_trace(Model.__qualname__, name)

            if type(field) is ListDataClassField:
                func = get_serialize_sequence(func)
            elif type(field) is DictDataClassField:
                func = get_serialize_dict(func)
            elif type(field) is Dict2ListDataClassField:
                func = get_serialize_dict(get_serialize_sequence(func))
        elif isinstance(field, (DateTimeField, DateField, TimestampField)):
            func = datetime_serializer
        else:
            logger.warning(
                "Generate serializer for model {}, but do not know how to serialize field {} yet. Using the default db_value from peewee",
                Model,
                name,
            )
            func = field.db_value
        field2func[name] = func

    def serialize_model(record: Model) -> dict:
        output = {}
        for name, func in field2func.items():
            value = getattr(record, name)
            if func is not None:
                output[name] = func(value)
            else:
                output[name] = value
        for name, name2 in foreign_keys:
            output[name] = getattr(record, name2)
        return output

    return serialize_model


def get_serializer_from_type(
    annotated_type, known_type_serializer: Dict[str, Serializer]
) -> Optional[Serializer]:
    if annotated_type in known_type_serializer:
        return known_type_serializer[annotated_type]
    if annotated_type in (str, int, float, bool, NoneType):
        return None
    if is_dataclass(annotated_type):
        return get_dataclass_serializer(annotated_type, known_type_serializer)
    if isinstance(annotated_type, _TypedDictMeta):
        # is_typeddict is not supported at python 3.8 yet
        return get_typeddict_serializer(annotated_type, known_type_serializer)

    try:
        if issubclass(annotated_type, Enum):
            return attrgetter("value")
    except TypeError:
        pass

    origin = get_origin(annotated_type)
    args = get_args(annotated_type)

    if origin is None or len(args) == 0:
        # we can't handle this type, e.g., some class that are not dataclass, or simply just list or set (not enough information)
        raise NoDerivedSerializer().add_trace(annotated_type)

    # handle literal first
    if origin is Literal:
        return None

    if (
        origin is list
        or origin is set
        or (origin is tuple and len(args) == 2 and args[1] is Ellipsis)
    ):
        ser_arg = get_serializer_from_type(args[0], known_type_serializer)
        if ser_arg is None:
            return None
        return get_serialize_sequence(ser_arg)

    if origin is dict:
        ser_value = get_serializer_from_type(args[1], known_type_serializer)
        if ser_value is None:
            return None
        return get_deserialize_dict(ser_value)

    if origin is Union:
        return get_serialize_union(
            classes=args,
            serializers=[
                get_serializer_from_type(arg, known_type_serializer) for arg in args
            ],
        )

    raise NoDerivedSerializer().add_trace(annotated_type)


def get_dataclass_serializer(
    CLS, known_type_serializer: Dict[str, Serializer]
) -> Serializer:
    field2serializer: Dict[str, Optional[Serializer]] = {}
    field_types = get_type_hints(CLS)

    def serialize_dataclass(obj):
        if obj is None:
            return None

        output = {}
        for field, serializer in field2serializer.items():
            value = getattr(obj, field)
            if serializer is None:
                output[field] = value
            else:
                output[field] = serializer(value)
        return output

    # assign first to support recursive type in the field
    known_type_serializer[CLS] = serialize_dataclass

    for field in fields(CLS):
        field_type = field_types[field.name]
        try:
            func = get_serializer_from_type(field_type, known_type_serializer)
        except NoDerivedSerializer as e:
            # can't automatically figure out how to serialize this field
            del known_type_serializer[CLS]
            raise e.add_trace(CLS.__qualname__, field.name)

        field2serializer[field.name] = func

    return serialize_dataclass


def get_typeddict_serializer(
    typeddict: _TypedDictMeta, known_type_serializer: Dict[str, Serializer]
) -> Serializer:
    total = typeddict.__total__
    if not total:
        # they can inject any key as the semantic of total, so we do not have serializer for this.
        raise NoDerivedSerializer().add_trace(
            typeddict.__name__, "is not a total TypedDict"
        )

    field2deserializer = {}

    def serialize_typeddict(value):
        if value is None:
            return None

        output = {}
        for field, deserializer in field2deserializer.items():
            value = getattr(typeddict, field)
            if deserializer is None:
                output[field] = value
            else:
                output[field] = deserializer(value)
        return output

    known_type_serializer[typeddict] = serialize_typeddict

    for field, field_type in typeddict.__annotations__.items():
        try:
            func = get_serializer_from_type(field_type, known_type_serializer)
        except NoDerivedSerializer as e:
            raise e.add_trace(field)
        field2deserializer[field] = func

    return serialize_typeddict


def get_serialize_sequence(serializer):
    def serialize_list(value):
        if value is None:
            return None
        return [serializer(item) for item in value]

    return serialize_list


def get_serialize_dict(serializer):
    def serialize_dict(value):
        if value is None:
            return None
        return {key: serializer(item) for key, item in value.items()}

    return serialize_dict


def get_serialize_union(classes, serializers):
    if all(ser is None for ser in serializers):
        return None

    def serialize_union(value):
        if value is None:
            return None
        for cls, serializer in zip(classes, serializers):
            try:
                if isinstance(value, cls):
                    return serializer(value)
            except:
                raise
        raise ValueError(
            f"Cannot serialize {value} which is instance of either of {classes}"
        )

    return serialize_union



from collections import defaultdict
from curses import raw
import re
from functools import partial
from typing import Mapping, Type, Callable, Any, List, Optional, Dict

from flask import Blueprint, json, request, jsonify
from gena.deserializer import generate_deserializer, NoDerivedDeserializer
from peewee import Model as PeeweeModel, DoesNotExist, fn
from playhouse.shortcuts import model_to_dict
from werkzeug.exceptions import BadRequest, NotFound

from gena.serializer import Serializer, get_peewee_serializer


def generate_api(
    Model: Type[PeeweeModel],
    deserializers: Optional[Dict[str, Callable[[Any], Any]]] = None,
    serialize: Optional[Callable[[Any], dict]] = None,
    batch_serialize: Optional[Callable[[List[Any]], List[dict]]] = None,
    known_type_serializer: Optional[Dict[Any, Serializer]] = None,
    enable_truncate_table: bool = False,
):
    """Generate API from the given Model

    Args:
        Model: peewee model of the table
        deserializers: deserialize raw value, throw value error if value is invalid. you can provide deserializer for some field and the rest is going to be generated automatically
        serialize: function to serialize a record of the table
        batch_serialize: function to serialize multiple records of the table
        known_type_serializer: serializer for known type, only affected if serialize is generated automatically
        enable_truncate_table: whether to enable API to truncate the whole table
    """
    table_name = Model._meta.table_name
    default_limit = str(50)
    name2field = {name: field for name, field in Model._meta.fields.items()}
    op_fields = {"fields", "limit", "offset", "unique", "sorted_by", "group_by"}
    field_reg = re.compile(r"(?P<name>[a-zA-Z_0-9]+)(?:\[(?P<op>[a-zA-Z0-9]+)\])?")

    if (
        deserializers is None
        or len(set(name2field.keys()).difference(deserializers.keys())) > 0
    ):
        if deserializers is None:
            deserializers = {}

        # automatically filling missing deserializers
        deserializers.update(
            **generate_deserializer(
                Model, known_field_deserializers=set(deserializers.keys())
            )
        )

    if len(set(name2field.keys()).difference(deserializers.keys())) > 0:
        raise Exception(
            f"Table {table_name} doesn't have deserializer for field: {set(name2field.keys()).difference(deserializers.keys())}"
        )

    if serialize is None:
        if hasattr(Model, "to_dict"):
            serialize = Model.to_dict
        elif batch_serialize is not None:
            serialize = lambda x: batch_serialize([x])[0]
        else:
            serialize = get_peewee_serializer(
                Model, known_type_serializer=known_type_serializer
            )

    if batch_serialize is None:
        assert serialize is not None
        batch_serialize = lambda lst: [serialize(item) for item in lst]

    assert len(op_fields.intersection(name2field.keys())) == 0

    bp = Blueprint(table_name, table_name)

    @bp.route(f"/{table_name}", methods=["GET"])
    def get():
        """Retrieving records matched a query.

        There are fixed keyword arguments:
            - limit: limit the number of records returned
            - offset: offset of the records returned
            - sorted_by: comma separated list of field name to sort by, prefix a field with "-" to sort that field in descending order.
            - group_by: comma separated list of field name to group by.
            - fields: comma separated list of field name to return. If not specified, all fields are returned.

        To filter records, you can apply the condition of a column by <field>=<value> (equal condition).
        Condition on a field such as >, >=, <, <=, `max`, `min`, `in` can be specified using brackets such as: <field>[gt]=10.
        We also support complex conditions:
            1. `max`: keep the record in a group that has the largest value
            2. `min`: keep the record in a group that has the smallest value
            3. `in`: select record that its values are in the given list

        The syntax for `max` and `min` are <field>[op]=<fields>, where op is either `max` or `min` and fields is a comma separated
        list of fields that will be used to group the records.

        We can support another aggregation to select keep all values in a group. However, since each value is for each
        record, we also have multiple ids. Therefore, a natural choice is to use `group_by` operator instead. Note that
        when you use group_by, the output is still a table (not a mapping) so that the client can reuse the code that
        read the data, but they have to group the result themselves.

        Note that we enforce the constraint that only one aggregation (`max`, `min`, `group_by`, etc) is allow in a query
        to ensure the behaviour of the query is deterministic (e.g., apply a group_by and max, which one is apply first?).
        """
        if "fields" in request.args:
            field_names = request.args["fields"].split(",")
            fields = [name2field[name] for name in field_names]
        else:
            field_names = []
            fields = []

        limit = int(request.args.get("limit", default_limit))
        offset = int(request.args.get("offset", "0"))

        # construct select clause
        query = Model.select(*fields)
        unique = request.args.get("unique", "false") == "true"
        if unique:
            query = query.distinct()

        # construct order by clause
        order_by = []
        if "sorted_by" in request.args:
            for field in request.args["sorted_by"].split(","):
                try:
                    if field.startswith("-"):
                        field = field[1:]
                        order_by.append(name2field[field].desc())
                    else:
                        order_by.append(name2field[field])
                except KeyError:
                    raise BadRequest(f"Invalid field name: {field}")

        if len(order_by) > 0:
            query = query.order_by(*order_by)

        # construct group by clause
        group_by = []
        if "group_by" in request.args:
            for field in request.args["group_by"].split(","):
                if field not in name2field:
                    raise BadRequest(f"Invalid field name: {field}")
                group_by.append(name2field[field])

        # construct where clause
        filter_fields = defaultdict(list)
        for name, value in request.args.items():
            if name in op_fields:
                continue
            m = field_reg.match(name)
            if m is None:
                raise BadRequest(f"Invalid field name: {name}")

            name = m.group("name")
            if name not in name2field:
                raise BadRequest(f"Invalid field name: {name}")

            op = m.group("op")
            filter_fields[name].append((op, value))

        pending_ops = defaultdict(dict)
        conditions = defaultdict(list)
        for name, ops in filter_fields.items():
            for op, value in ops:
                field = name2field[name]
                if op in {"max"}:
                    assert op not in pending_ops[name]
                    pending_ops[name][op] = value
                    continue
                elif op == "in":
                    deser = deserializers[name]
                    conditions[field].append(
                        (field.in_([deser(x) for x in value.split(",")]))
                    )
                    continue

                # no special operator
                value = deserializers[name](value)
                if op is None:
                    conditions[field].append((field == value))
                elif op == "gt":
                    conditions[field].append((field > value))
                elif op == "gte":
                    conditions[field].append((field >= value))
                elif op == "lt":
                    conditions[field].append((field < value))
                elif op == "lte":
                    conditions[field].append((field <= value))
                else:
                    raise BadRequest(f"Does not support {op} yet")

        if len(conditions) > 0:
            query = query.where(*[item for lst in conditions.values() for item in lst])

        if len(group_by) > 0:
            if len(pending_ops) > 0:
                raise BadRequest(f"Does not support multiple aggregations")
            # update the select to keep the id
            subquery = query.select(
                *[c.alias(f"gb_c{i}") for i, c in enumerate(group_by)]
            ).group_by(*group_by)

            predicate = group_by[0] == getattr(subquery.c, "gb_c0")
            for i, c in enumerate(group_by[1:], start=1):
                predicate = predicate & (c == getattr(subquery.c, f"gb_c{i}"))
            query = query.join(subquery.limit(limit).offset(offset), on=predicate)
            # they want to get only one record so we save computation knowing that it won't use anyway
            total = subquery.count()
        else:
            for name, ops in pending_ops.items():
                field = name2field[name]
                for op, value in ops.items():
                    if op == "max":
                        # select the record with maximum value in the group
                        # need to do a subquery to select the one with maximum value
                        subquery_group_fields = []
                        subquery_group_field_conditions = []

                        for gfield in value.split(","):
                            if gfield not in name2field:
                                raise BadRequest(f"Invalid group by field: {gfield}")
                            subquery_group_fields.append(name2field[gfield])
                            if gfield in conditions:
                                if gfield in pending_ops:
                                    raise BadRequest(
                                        f"Does not support multiple aggregations"
                                    )
                                subquery_group_field_conditions += conditions[gfield]

                        subquery_name = f"{name}_{op}"
                        field_alias = f"{subquery_name}_{name}"
                        subquery = (
                            Model.select(Model.id, fn.MAX(field).alias(field_alias))
                            .group_by(*subquery_group_fields)
                            .alias(subquery_name)
                        )

                        if len(subquery_group_field_conditions) > 0:
                            subquery = subquery.where(*subquery_group_field_conditions)

                        predicate = (Model.id == subquery.c.id) & (
                            field == getattr(subquery.c, field_alias)
                        )
                        query = query.join(subquery, on=predicate)

            # they want to get only one record so we save computation knowing that it won't use anyway
            total = query.count()
            query = query.limit(limit).offset(offset)

        # perform the query
        items = batch_serialize(list(query))
        if len(fields) > 0:
            items = [{k: item[k] for k in field_names} for item in items]

        return jsonify({"items": items, "total": total})

    @bp.route(f"/{table_name}/find_by_ids", methods=["POST"])
    def get_by_ids():
        if "ids" not in request.json:
            raise BadRequest("Bad request. Missing `ids`")

        ids = request.json["ids"]
        if "fields" in request.args:
            field_names = request.args["fields"].split(",")
            fields = [name2field[name] for name in field_names]
        else:
            field_names = []
            fields = []

        records = list(Model.select(Model.id, *fields).where(Model.id.in_(ids)))
        records = batch_serialize(records)

        if len(field_names) > 0:
            records = {
                item["id"]: {k: item[k] for k in field_names if k in item}
                for item in records
            }

        return jsonify({"items": records, "total": len(records)})

    @bp.route(f"/{table_name}/<id>", methods=["GET"])
    def get_one(id):
        if "fields" in request.args:
            field_names = request.args["fields"].split(",")
            fields = [name2field[name] for name in field_names]
        else:
            field_names = []
            fields = []

        records = list(Model.select(*fields).where(Model.id == id))
        if len(records) == 0:
            raise NotFound(f"Record {id} does not exist")

        record = serialize(records[0])
        if len(fields) > 0:
            record = {k: record[k] for k in field_names}

        return jsonify(record)

    @bp.route(f"/{table_name}/<id>", methods=["HEAD"])
    def has(id):
        if not Model.select().where(Model.id == id).exists():
            raise NotFound(f"Record {id} does not exist")
        return jsonify()

    @bp.route(f"/{table_name}", methods=["POST"])
    def create():
        posted_record = request.json
        raw_record = {}

        for name, field in name2field.items():
            if name in posted_record:
                try:
                    raw_record[name] = deserializers[name](posted_record[name])
                except ValueError as e:
                    raise ValueError(f"Field `{name}` {str(e)}")
        if "id" in raw_record:
            # remove id as this API always creates a new record
            raw_record.pop("id")
        record = Model.create(**raw_record)
        # TODO: correct return types according to RESTful specification https://restfulapi.net/http-methods/
        return jsonify(serialize(record))

    @bp.route(f"/{table_name}/<id>", methods=["PUT"])
    def update(id):
        try:
            record = Model.get_by_id(id)
        except DoesNotExist as e:
            raise NotFound(f"Record {id} does not exist")

        request_json = request.get_json()
        if request_json is None:
            raise BadRequest("Missing request body")

        for name, field in name2field.items():
            if name in request_json:
                try:
                    value = deserializers[name](request_json[name])
                except ValueError as e:
                    raise ValueError(f"Field `{name}` {str(e)}")

                setattr(record, name, value)
        record.save()

        return jsonify(serialize(record))

    @bp.route(f"/{table_name}/<id>", methods=["DELETE"])
    def delete_by_id(id):
        try:
            Model.get_by_id(id).delete_instance()
        except DoesNotExist as e:
            raise NotFound(f"Record {id} does not exist")

        return jsonify({"status": "success"})

    if enable_truncate_table:

        @bp.route(f"/{table_name}", methods=["DELETE"])
        def truncate():
            Model.truncate_table()
            return jsonify({"status": "success"})

    return bp


def generate_readonly_api_4dict(
    name: str,
    id2ent: Mapping[str, Any],
    unique_field_funcs: Dict[str, Callable[[str], str]] = None,
    serialize: Optional[Callable[[Any], dict]] = None,
    batch_serialize: Optional[Callable[[List[Any]], List[dict]]] = None,
):
    """Generate API for a dictionary.

    Make it as similar to the API of DB as possible. However, due to its limitation, we
    only support accessing by unique key (via id or other unique field)

    Args:
        name: name of the endpoint
        id2ent: dictionary of entities
        unique_field_funcs: unique field and function to transform it to id
        serialize: function to serialize an entity
        batch_serialize: function to serialize a list of entities
    """
    op_fields = {"fields", "limit", "offset", "unique", "sorted_by"}
    if batch_serialize is None:
        assert serialize is not None
        batch_serialize = gen_batch_serialize(serialize)
    elif serialize is None:
        serialize = lambda x: batch_serialize([x])[0]

    if unique_field_funcs is None:
        unique_field_funcs = {}
    bp = Blueprint(name, name)

    @bp.route(f"/{name}", methods=["GET"])
    def get():
        """Retrieving records matched a query."""
        if "fields" in request.args:
            field_names = request.args["fields"].split(",")
        else:
            field_names = []

        lst = []
        for name in request.args.keys():
            if name in op_fields:
                continue
            if name not in unique_field_funcs:
                raise BadRequest(f"Invalid field name: {name}")
            lst.append(name)

        if len(lst) > 1:
            raise BadRequest(f"Invalid query. Only one field is allowed but get: {lst}")

        if len(lst) == 0:
            raise BadRequest(f"Invalid query. Must provide at least one field")

        id = unique_field_funcs[lst[0]](request.args[lst[0]])
        record = serialize(id2ent[id])
        if len(field_names) > 0:
            record = {k: record[k] for k in field_names if k in record}

        return jsonify({"items": [record], "total": 1})

    @bp.route(f"/{name}/find_by_ids", methods=["POST"])
    def find_by_ids():
        if "ids" not in request.json:  # type: ignore
            raise BadRequest("Bad request. Missing `ids`")

        if "fields" in request.args:
            field_names = request.args["fields"].split(",")
        else:
            field_names = []

        ids = []
        ents = []
        for id in request.json["ids"]:
            if id in id2ent:
                ents.append(id2ent[id])
                ids.append(id)

        records = batch_serialize(ents)
        if len(field_names) > 0:
            records = [
                {k: item[k] for k in field_names if k in item} for item in records
            ]

        return jsonify({"items": dict(zip(ids, records)), "total": len(ents)})

    @bp.route(f"/{name}/<id>", methods=["GET"])
    def find_by_id(id: str):
        if id not in id2ent:
            raise NotFound(f"Record {id} does not exist")

        record = serialize(id2ent[id])

        if "fields" in request.args:
            field_names = request.args["fields"].split(",")
        else:
            field_names = []

        if len(field_names) > 0:
            record = {k: record[k] for k in field_names if k in record}

        return jsonify(record)

    return bp


def gen_batch_serialize(
    serialize: Callable[[Any], dict]
) -> Callable[[List[Any]], List[dict]]:
    def batch_serialize(lst):
        return [serialize(item) for item in lst]

    return batch_serialize



import importlib
import logging
import os
import pkgutil
from modulefinder import Module
from pathlib import Path
from typing import Optional, Union, List

from flask import Flask, render_template, Blueprint, send_from_directory


def generate_app(
    controllers: Union[List[Blueprint], Module],
    pkg_dir: Union[str, Path],
    log_sql_queries: bool = True,
) -> Flask:
    if log_sql_queries and any(
        os.environ.get(key, "") == "development" for key in ["FLASK_ENV", "FLASK_DEBUG"]
    ):
        # if debugging, log the SQL queries
        logger = logging.getLogger("peewee")
        logger.addHandler(logging.StreamHandler())
        logger.setLevel(logging.DEBUG)

    app = Flask(
        __name__,
        template_folder=os.path.join(pkg_dir, "www"),
        static_folder=os.path.join(pkg_dir, "www/static"),
        static_url_path="/static",
    )
    app.config["app.json.sort_keys"] = False

    @app.route("/", defaults={"path": ""})
    @app.route("/<path:path>")
    def home(path):
        if path.find("/") == -1 and path.find(".") != -1:
            if path.endswith(".json") or path.endswith(".ico") or path.endswith(".png"):
                return send_from_directory(app.template_folder, path)

        return render_template("index.html")

    if isinstance(controllers, list):
        blueprints = controllers
    else:
        # auto discover blueprints
        blueprints = []
        for m in pkgutil.iter_modules(controllers.__path__):
            controller = importlib.import_module(f"{controllers.__name__}.{m.name}")
            for attrname in dir(controller):
                attr = getattr(controller, attrname)
                if isinstance(attr, Blueprint):
                    blueprints.append(attr)

    for bp in blueprints:
        app.register_blueprint(bp, url_prefix="/api")

    return app



from __future__ import annotations
from typing import Type
from flask import Flask
from flask.testing import FlaskClient
import pytest
from peewee import Model


class APITestSuite:
    @pytest.fixture
    def app(self) -> Flask:
        """Flask application"""
        raise NotImplementedError()

    @pytest.fixture
    def client(self, app: Flask) -> FlaskClient:
        """A test client returned by the Flask application"""
        return app.test_client()

    @pytest.fixture(scope="session")
    def model(self) -> Type[Model]:
        """Return the peewee model representing a database table holding the API resources"""
        raise NotImplementedError()

    @pytest.fixture(scope="session")
    def existed_resources(self) -> list[tuple[Model, dict]]:
        """Return a list of resources that already exist in the database (does not have to be exhaustive)"""
        raise NotImplementedError()

    @pytest.fixture(scope="session")
    def new_resources(self) -> list[dict]:
        """Return a list of new resources to be created"""
        raise NotImplementedError()

    def get_api_prefix(self, model: Type[Model]) -> str:
        return f"/api/{model._meta.table_name}"

    def test_get(
        self,
        client: FlaskClient,
        model: Type[Model],
        existed_resources: list[tuple[Model, dict]],
    ):
        # test getting the existed resources.
        # Note: the existed resources must be within top 100
        result = client.get(
            self.get_api_prefix(model), query_string={"limit": 100, "offset": 0}
        )

        returned_resources = {x["id"]: x for x in result.json["items"]}

        for resource, resource_json in existed_resources:
            rid = resource.id
            assert rid in returned_resources
            assert returned_resources[rid] == resource_json

    def test_create(
        self, client: FlaskClient, model: Type[Model], new_resources: list[dict]
    ):
        for new_resource in new_resources:
            resp = client.post(self.get_api_prefix(model), json=new_resource)
            assert resp.status_code == 200



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gena']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.2.2,<3.0.0',
 'loguru>=0.6.0,<0.7.0',
 'peewee>=3.15.2,<4.0.0',
 'requests>=2.28.0,<3.0.0',
 'typing_extensions>=4.0.0,<5.0.0']

setup_kwargs = {
    'name': 'gena',
    'version': '1.4.1',
    'description': 'Generate APIs from Database Models',
    'long_description': '# GENA\n\nFramework to help to build (web) application faster.\n\n1. [`gena`](): generate restful APIs from peewee models (i.e., specification of database table).\n2. [`gena-app`](): provides basic structure for frontend application (state management, routing).\n\nFor demo, see the [todolist folder](/examples/todolist).\n\n# Getting started\n\nWe are going to build a simple todolist app to demonstrate how to use this framework. See the [todolist folder](/todolist) for the complete source code.\n\n### Setup the project:\n\n1. create the python project (server code): `poetry new todolist` (you need [Poetry](https://python-poetry.org/) to run this command)\n2. move inside the project folder: `cd todolist`\n3. add our backend library: `poetry add gena`\n4. create a `www` directory containing your frontend code using [`create-react-app`](https://create-react-app.dev/docs/adding-typescript/): (`yarn create react-app www --template typescript`)\n5. add our frontend library: `cd www; yarn add gena-app; cd ..` -- this will modify your [`www/package.json`](/todolist/www/package.json) file.\n6. modify the build script in [`www/package.json`](/todolist/www/package.json) to tell `create-react-app` to build static files into [`todolist/www`](/todolist/todolist/www) directory inside the python package: `"build": "BUILD_PATH=\'../todolist/www\' react-scripts build"`. This allows us to distribute both frontend and backend in a single python package.\n7. add `"proxy": "http://localhost:5000"` to the [`www/package.json`](/todolist/www/package.json). This allows us to send the request to the server during development.\n\nAfter this step, you will have the following folder structure:\n\n    todolist\n    ├── todolist                   # our backend code\n    │   ├── __init__.py\n    ├── pyproject.toml\n    ├── www                        # our frontend code\n\n### Backend\n\n#### Define database schema\n\nWe use [`peewee`](https://docs.peewee-orm.com/), an ORM library, to define our database schema.\n\nFirst, let\'s create a file containing our schema: [`todolist/models.py`](/todolist/todolist/models.py). In this simple todo list, we only have one table `todo_list`, which has two fields: `checked` and `todo`.\n\n```python\nimport os\nfrom peewee import SqliteDatabase, Model, BooleanField, TextField\n\n\ndb = SqliteDatabase(os.environ["DBFILE"])\n\n\nclass TodoList(Model):\n    class Meta:\n        database = db\n        db_table = "todo_list"\n\n    checked = BooleanField()\n    todo = TextField()\n```\n\n(Optionally) We add the following code to the same file to insert some dummy data to our database for testing\n\n```python\nif not os.path.exists(os.environ[\'DBFILE\']):\n    db.create_tables([TodoList], safe=True)\n    TodoList.insert_many([\n        {"checked": False, "todo": "go grocery"},\n        {"checked": False, "todo": "do laundry"},\n    ]).execute()\n```\n\n#### Create APIs\n\nWith our backend library, creating APIs is as simple as calling a function `generate_api` with the ORM model `TodoList`:\n\n```python\nimport os\nfrom gena import generate_app, generate_api\nfrom todolist.models import TodoList\n\napp = generate_app(\n    controllers=[generate_api(model) for model in [TodoList]],\n    pkg_dir=os.path.dirname(__file__)\n)\n```\n\nUnder the hood, [`generate_api`](/gena/api_generator.py) uses the specification in the model to automatically generate a blueprint containing the following endpoints (sometimes are called views or controllers):\n\n1. `GET /{table_name}`: querying records matched a query.\n2. `GET /{table_name}/<id>`: get a record by id\n3. `HEAD /{table_name}/<id>`: whether a record with the given id exists.\n4. `POST /{table_name}/find_by_ids`: querying list of records by their ids\n5. `POST /{table_name}`: create a record\n6. `PUT /{table_name}/<id>`: update a record\n7. `DELETE /{table_name}/<id>`: delete a record\n8. `DELETE /{table_name}`: truncate the whole table, only available if the `enable_truncate_table` parameter is set to be `True`.\n\nYou can extend the blueprint to add additional endpoints or overwrite some if you needed. Inputs from the users are automatically validate based on the type of the field, or you can provide your own deserializer (whose job is also validate the value) for a particular field via the `deserializer` parameter. You can also provide your own `serializer` to control how the data should be serialize.\n\nThe outer function `generate_app` will returns the Flask application. It takes three parameters\n\n- `controllers`: a list of [Flask\'s blueprints](https://flask.palletsprojects.com/en/2.0.x/blueprints/) or a python package, in which the blueprints are discovered automatically. This is useful if you organize your blueprints in separated files. All endpoints defined in the blueprints have url prefix `/api` (`/{table_name}` becomes `/api/{table_name}`).\n- `pkg_dir`: the path to our `todolist` package. Our app will serve the static files from `{pkg_dir}/www` folder.\n- `log_sql_queries`: whether to log sql queries when we are in development mode.\n\nWith the given app, we finally can start our server with: `DBFILE=./todolist.db FLASK_APP=todolist/app.py flask run`. Note: remember to activate your virtual environment first with `poetry shell`.\n\n### Frontend\n\n#### Create Stores\n\nOur frontend library uses [React](https://reactjs.org/) and provides state management with [MobX](https://mobx.js.org/)) and routing with [ReactRouter](https://reactrouter.com/) out of the box. In addition, most of the functions are typed so that refactoring and maintaining is easier.\n\nTo start, we create stores which contains all of the data of our application. A store also contains other functions to create, update, delete, and query records from the server. If you haven\'t familiar with MobX, please check out their [documentation](https://mobx.js.org/the-gist-of-mobx.html).\n\nLet\'s create a file [`www/src/models/TodoListStore.ts`](/todolist/www/src/models/TodoListStore.ts) and paste the following code:\n\n```typescript\nimport { SimpleCRUDStore } from "gena-app";\nimport { makeObservable, action } from "mobx";\n\nexport interface Todo {\n  id: number;\n  checked: boolean;\n  todo: string;\n}\n\nexport class TodoListStore extends SimpleCRUDStore<number, Todo> {\n  constructor() {\n    super(`/api/todo_list`);\n\n    makeObservable(this, {\n      toggle: action,\n    });\n  }\n\n  toggle(item: Todo) {\n    item.checked = !item.checked;\n  }\n}\n```\n\nIn the above code, we create an interface `Todo` matched with the definition in our database schema. Since we didn\'t provide custom deserializer, the fields in our frontend is matched exactly with the fields in our database in the backend. Then, we define a store `TodoListStore` extending the `SimpleCRUDStore` from our frontend library `gena-app`. This gives us several useful functions out of the box such as `create`, `delete`, `update`, and `query` the records from the server. We also define an additional function that will toggle the `checked` value of an `Todo` to show how to extend the store if needed.\n\nTo use the stores with [React Hooks](https://reactjs.org/docs/hooks-intro.html), we need to following code in [`www/src/models/index.ts`](/todolist/www/src/models/index.ts)\n\n```typescript\nimport React from "react";\nimport { TodoListStore } from "./TodoListStore";\n\nexport const stores = {\n  todolistStore: new TodoListStore(),\n};\nexport type IStore = Readonly<typeof stores>;\nexport const StoreContext = React.createContext<IStore>(stores);\n\nexport function useStores(): IStore {\n  return React.useContext(StoreContext);\n}\n\nexport { TodoListStore as VariableStore };\nexport type { Todo } from "./TodoListStore";\n```\n\n#### Routing & Application\n\nRouting specifies which component the application should render for a particular URL. This makes it much easier to keep the application logic separated especially when you have lots of pages. Our frontend library enables us to define all the routing in a single place with strong type annotation, so that it\'s easier to maintain the application\'s routing and is harder to make mistake.\n\nLet\'s create a file `www/routes.tsx` and define our first route:\n\n```typescript\nimport { NoArgsPathDef } from "gena-app";\nimport { HomePage } from "./pages/HomePage";\n\nexport const routes = {\n  home: new NoArgsPathDef({\n    component: HomePage,\n    pathDef: "/",\n    exact: true,\n  }),\n};\n```\n\nIn this route, we say when user open `/`, we will render a component [`HomePage`](/todolist/www/src/pages/HomePage.tsx). For now, let\'s just use a dummy `HomePage` for now:\n\n```typescript\nexport const HomePage = () => {\n  return <p>Hello world</p>;\n};\n```\n\nWith the specified routes and a dummy `HomePage` component, the final piece is to render and attach our component to the DOM. In [`index.tsx`](/todolist/www/src/index.tsx):\n\n```typescript\nimport ReactDOM from "react-dom";\nimport { App } from "gena-app";\nimport "./index.css";\nimport { StoreContext, stores } from "./models";\nimport reportWebVitals from "./reportWebVitals";\nimport { routes } from "./routes";\n\nReactDOM.render(\n  <StoreContext.Provider value={stores}>\n    <App routes={routes} />\n  </StoreContext.Provider>,\n  document.getElementById("root")\n);\n```\n\n`StoreContext.Provider` component enables us to use `useStores` hook we defined earlier in any component, and `App` component will render the component that matched with the current URL as specified in our routes.\n\n#### Create TodoList component\n\nIt\'s now time to complete our [`HomePage`](/todolist/www/src/pages/HomePage.tsx) component to show the TodoList.\n\nWe first use the stores within the component and wrapped it with `observer` so that it will react whenever the state changes\n\n```typescript\nimport { observer } from "mobx-react";\n\nexport const HomePage = observer(() => {\n  const { todolistStore } = useStores();\n  ...\n});\n```\n\nWe are going to display the todo list as a list. So we use [List component](https://ant.design/components/list).\n\n```typescript\nexport const HomePage = observer(() => {\n  const { todolistStore } = useStores();\n  useEffect(() => {\n    todolistStore.fetch({ limit: 1000, offset: 0 });\n  }, []);\n\n  const items = todolistStore.list.map((item) => {\n    return (\n      <List.Item key={item.id}>\n        <Checkbox\n          checked={item.checked}\n          onChange={(e) => {\n            item.checked = e.target.checked;\n            todolistStore.update(item);\n          }}\n        >\n          <Typography.Paragraph\n            style={{ marginBottom: 0 }}\n            editable={{\n              onChange: (text) => {\n                item.todo = text;\n                todolistStore.update(item);\n              },\n            }}\n          >\n            {item.todo}\n          </Typography.Paragraph>\n        </Checkbox>\n        <Button\n          type="primary"\n          danger={true}\n          onClick={() => {\n            todolistStore.delete(item.id);\n          }}\n        >\n          Delete\n        </Button>\n      </List.Item>\n    );\n  });\n\n  const addItem = () => todolistStore.create({ checked: false, todo: "" });\n\n  return (\n    <Space direction="vertical" style={{ width: "100%" }}>\n      <List bordered={true}>{items}</List>\n      <Button type="primary" onClick={addItem}>\n        Add\n      </Button>\n    </Space>\n  );\n});\n```\n\nNow, we are able to display the list of todo items from the database, and check/uncheck item that is done. We can extend the UI to add other functionalities such as add, delete, or update the content of todo items. You can check out the complete code in [here](/todolist/www/src/pages/HomePage.tsx).\n',
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
